/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.netDesignAlgorithm.fa;

import cern.colt.matrix.tdouble.DoubleMatrix2D;
import com.jom.OptimizationProblem;
import com.net2plan.interfaces.networkDesign.IAlgorithm;
import com.net2plan.interfaces.networkDesign.Net2PlanException;
import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.libraries.GraphUtils;
import com.net2plan.utils.Constants;
import com.net2plan.utils.DoubleUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Given a network topology, the capacities in the links and the offered traffic, 
 * this algorithm obtains the destination-based traffic routing that minimizes the average number 
 * of hops. Destination-based routing means that the routing could be implemented i.e. in an IP network.
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.2, September 2015
 */
public class FA_minAvNumHops_xte implements IAlgorithm
{
	@Override
	public String executeAlgorithm(NetPlan netPlan, Map<String, String> algorithmParameters, Map<String, String> net2planParameters)
	{
		/* Basic checks */
		int N = netPlan.getNumberOfNodes();
		int E = netPlan.getNumberOfLinks();
		int D = netPlan.getNumberOfDemands();
		if (N == 0 || D == 0 || E == 0)	throw new Net2PlanException("This algorithm requires a topology with links, and a demand set");
		
		/* Initialize some variables */
		double[] u_e = netPlan.getLinkCapacityVector();
		double[] h_d = netPlan.getDemandOfferedTrafficVector();
		DoubleMatrix2D trafficMatrix = netPlan.getTrafficMatrix();
		
		/* The diagonal in the traffic matrix contains minus the amount of traffic destined to that node */
		for (int n1 = 0; n1 < N; n1++)
			for (int n2 = 0; n2 < N; n2++)
				if (n1 != n2)
					trafficMatrix.setQuick(n2, n2, trafficMatrix.getQuick(n2, n2) - trafficMatrix.getQuick(n1, n2));

		/* Remove all routes in current netPlan object */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);

		/* Create the optimization problem object (JOM library) */
		OptimizationProblem op = new OptimizationProblem();

		/* Add the decision variables to the problem */
		op.addDecisionVariable("x_te", false, new int[] {N, E}, 0, Double.MAX_VALUE);

		/* Set some input parameters */
		op.setInputParameter("u_e", u_e, "row");
		op.setInputParameter("H_d", DoubleUtils.sum(h_d));

		/* Sets the objective function */
		op.setObjectiveFunction("minimize", "sum(sum(x_te)) / H_d");

		/* VECTORIAL FORM OF THE CONSTRAINTS [EQUIVALENT TO THE FOR-LOOP FORM, COMMENT ONE OF BOTH] */
		Set<Long> nodeIds = netPlan.getNodeIds();
		Map<Long, Pair<Long, Long>> linkMap = netPlan.getLinkMap();
		op.setInputParameter("A_ne", GraphUtils.getIncidenceMatrix(nodeIds, linkMap));
		op.setInputParameter("TM", trafficMatrix);
		op.addConstraint("A_ne * x_te' == TM"); /* the flow-conservation constraints (NxN constraints) */
		op.addConstraint("sum(x_te,1) <= u_e"); /* the capacity constraints (E constraints) */

		/* FOR-LOOP BASED FORM OF THE CONSTRAINTS [EQUIVALENT TO THE VECTORIAL FORM, COMMENT ONE OF BOTH] */
//		/* Flow conservation constraints */
//		for (int n = 0 ; n < N ; n ++)
//		{
//			int [] outLinks = netPlan.getNodeOutgoingLinks(n);
//			int [] inLinks = netPlan.getNodeIncomingLinks(n);
//			op.setInputParameter("delta_out", outLinks , "row");
//			op.setInputParameter("delta_in", inLinks , "row");
//			for (int t = 0 ; t < N ; t ++)
//			{
//				op.setInputParameter("t", t);
//				op.addConstraint ("sum(x_te(t,delta_out)) - sum(x_te(t,delta_in)) == " + trafficMatrix[n][t]);
//			}
//		}
//		/* Capacity - min congestion constraints */
//		for (int e = 0 ; e < E ; e ++)
//			op.addConstraint("sum(x_te(all," + e + ")) <= " + u_e[e] + " - u");

		/* Call the solver to solve the problem */
		String solverName = algorithmParameters.get("solverName");
		String solverLibraryName = algorithmParameters.get("solverLibraryName");
		op.solve(solverName, "solverLibraryName", solverLibraryName);
		
		/* If an optimal solution was not found, quit */
		if (!op.solutionIsOptimal()) throw new Net2PlanException("An optimal solution was not found");

		/* Retrieve the optimum solutions */
		DoubleMatrix2D x_te = op.getPrimalSolution("x_te").view2D();

		/* Update netPlan object: routing */
		netPlan.removeAllRoutingInformation();
		netPlan.setRoutingType(Constants.RoutingType.SOURCE_ROUTING);
		netPlan.addRoutesFromDestinationLinkCarriedTraffic(x_te);

		return "Ok!";
	}

	@Override
	public String getDescription()
	{
		return "Given a network topology, the capacities in the links and the "
			+ "offered traffic, this algorithm obtains the destination-based "
			+ "traffic routing that minimizes the average number of hops. "
			+ "Destination-based routing means that the routing could be "
			+ "implemented i.e. in an IP network";
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("solverName", "glpk", "The solver name to be used by JOM"));
		algorithmParameters.add(Triple.of("solverLibraryName", "", "The solver library full or relative path, to be used by JOM. Leave blank to use JOM default."));
		
		return algorithmParameters;
	}
}